#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2008, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#


# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0


# name reference test #001
# Note we run this test in a seperate shell to make sure the memory
# corruption originally reported can be reproduced (which precisely
# depends on ordering in the testcase)
function test1
{
{
cat <<EOF
	function err_exit
	{
		print -u2 -n '\t'
		print -u2 -r \${Command}[\$1]: "\${@:2}"
		(( Errors++ ))
	}
	alias err_exit='err_exit \$LINENO'

	function function2
	{
		nameref v=\$1

		v.x=19
		v.y=20
	}

	function function1
	{
		typeset compound_var=()

		function2 compound_var

		printf "x=%d, y=%d\n" compound_var.x compound_var.y 
	}

	x="\$(function1)"

	[[ "\$x" == 'x=19, y=20' ]] || err_exit "expected 'x=19, y=20', got '\${x}'"

EOF
} | ${SHELL}
	(( Errors+=$? ))
	return 0
}

# name reference test #002
# Originally derived from the xmldocumenttree1.sh demo which failed
# with ast-ksh.2009-04-15 since the nodepath+nodesnum nameref calls
# were removing the compound variable members nodes+nodesnum (caused
# by a scoping bug)
function test2
{
	compound tmp
	typeset out
	integer numlines
{
cat <<EOF
	compound xdoc
	compound -A xdoc.nodes
	integer xdoc.nodesnum=0

	function test1
	{
        	nameref doc='xdoc'
        	nameref nodepath='doc.nodes'
        	nameref nodesnum='doc.nodesnum'
        	print -v doc
	}

	test1
EOF
} | out=$( ${SHELL} 2>&1 ) || err_exit "shell returned exit code $?"

	numlines="${ wc -l <<<"${out}" ; }"
	(( numlines == 4 )) || err_exit "Expected four lines of output, got ${out}"
	(set -o errexit ; read -C tmp <<<"${out}" ; [[ "$(typeset +p tmp.nodes)" == ~(Alr)(.*-A.*)&(.*-C.*) ]]) || err_exit 'missing variable tmp.nodes'
	(set -o errexit ; read -C tmp <<<"${out}" ; [[ "$(typeset +p tmp.nodesnum)" == ~(Alr)(.*-i.*)&(.*-l.*) ]]) || err_exit 'missing variable tmp.nodesnum'
	(set -o errexit ; read -C tmp <<<"${out}" ; [[ -v tmp.nodesnum ]]) || err_exit 'missing variable tmp.nodesnum'

	return 0
}


# name reference test #003a
# ast-ksh.2009-06-30 failed with the following compound variable/nameref test
function test3a
{
	typeset out
{
cat <<EOF
	compound -A addrsp

	nameref sp=addrsp
        
	sp[14]=( size=1 )
        
	if [[ -v sp[19] ]] ; then
        	print 'should not happen'
	else
        	print 'Ok'
	fi
EOF
} | out=$( ${SHELL} 2>&1 ) || err_exit "shell returned exit code $?"
	[[ "${out}" == 'Ok' ]] || err_exit "Expected 'Ok', got ${out}"
	return 0
}

# name reference test #003b
# (same as test #003a but uses a function)
# ast-ksh.2009-06-30 failed with the following compound variable/nameref test
function test3b
{
	typeset out
{
cat <<EOF
	compound -A addrsp
	
	function t1
	{
		nameref sp=\$1
        
		sp[14]=( size=1 )
        
		if [[ -v sp[19] ]] ; then
        		print "should not happen"
		else
        		print "Ok"
		fi
	}

	t1 addrsp
EOF
} | out=$( ${SHELL} 2>&1 ) || err_exit "shell returned exit code $?"
	[[ "${out}" == "Ok" ]] || err_exit "Expected 'Ok', got ${out}"
	return 0
}

# name reference test #004a
# (same as #003a but uses an indexed array instead of an associative one)
# ast-ksh.2009-06-30 failed with the following compound variable/nameref test
function test4a
{
	typeset out
{
cat <<EOF
	compound -a addrsp

	nameref sp=addrsp
        
	sp[14]=( size=1 )
        
	if [[ -v sp[19] ]] ; then
        	print 'should not happen'
	else
        	print 'Ok'
	fi
EOF
} | out=$( ${SHELL} 2>&1 ) || err_exit "shell returned exit code $?"
	[[ "${out}" == 'Ok' ]] || err_exit "Expected 'Ok', got ${out}"
	return 0
}


# name reference test #004b
# (same as test #004a but uses a function)
# ast-ksh.2009-06-30 failed with the following compound variable/nameref test
function test4b
{
	typeset out
{
cat <<EOF
	compound -a addrsp
	
	function t1
	{
		nameref sp=\$1
        
		sp[14]=( size=1 )
        
		if [[ -v sp[19] ]] ; then
        		print 'should not happen'
		else
        		print 'Ok'
		fi
	}

	t1 addrsp
EOF
} | out=$( ${SHELL} 2>&1 ) || err_exit "shell returned exit code $?"
	[[ "${out}" == 'Ok' ]] || err_exit "Expected 'Ok', got ${out}"
	return 0
}

# test for array type mutation 1
# $ ~/bin/ksh -c 'compound c=( typeset -a x ) ; nameref x=c.x ; x[4]=1 ;
# [[ ${ typeset -p c.x ;} != *-C* ]] || print mutation'
# -- snip --
# ast-ksh.2010-08-26 printed "mutation" in this case
function test_mutation1
{
	compound out=( typeset stdout stderr ; integer res )
	typeset testscript
	typeset -r expected_output='OK1'
	
	typeset -r -a tests=(
		# original testcase
		'compound c=( typeset -a x ) ; nameref x=c.x ; x[4]=1 ; if [[ ${ typeset -p c.x ;} != *-C* && ${ typeset -p c.x ;} == *-a* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		# same as original testcase but "x" is an associative array instead
		'compound c=( typeset -A x ) ; nameref x=c.x ; x[4]=1 ; if [[ ${ typeset -p c.x ;} != *-C* && ${ typeset -p c.x ;} == *-A* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		# same as original but "c" is now an { indexed, associative } array
		'compound -a c=( [2]=( typeset -a x ) ) ; nameref x=c[2].x ; x[4]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-a* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -a x ) ) ; nameref x=c[2].x ; x[4]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-a* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -A x ) ) ; nameref x=c[2].x ; x[4]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-A* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -A x ) ) ; nameref x=c[2].x ; x[4]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-A* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -a x ) ) ; nameref x=c[2].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -a x ) ) ; nameref x=c[2].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -A x ) ) ; nameref x=c[2].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -A x ) ) ; nameref x=c[2].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2].x ;} != *-C* && ${ typeset -p c[2].x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2][3]=( typeset -a x ) ) ; nameref x=c[2][3].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2][3].x ;} != *-C* && ${ typeset -p c[2][3].x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2][3]=( typeset -A x ) ) ; nameref x=c[2][3].x ; x[4][5]=1 ; if [[ ${ typeset -p c[2][3].x ;} != *-C* && ${ typeset -p c[2][3].x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		# same as previous test but use a new nameref c2 for c[2]
		'compound -a c=( [2]=( typeset -a x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-a* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -a x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-a* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -A x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-A* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -A x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-A* ]] ; then print "OK${x[4]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -a x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -a x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2]=( typeset -A x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -A c=( [2]=( typeset -A x ) ) ; nameref c2=c[2] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2][3]=( typeset -a x ) ) ; nameref c2=c[2][3] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-a* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
		'compound -a c=( [2][3]=( typeset -A x ) ) ; nameref c2=c[2][3] ; nameref x=c2.x ; x[4][5]=1 ; if [[ ${ typeset -p c2.x ;} != *-C* && ${ typeset -p c2.x ;} == *-A* ]] ; then print "OK${x[4][5]}" ; else print mutation ; fi'
	)
	integer ti # tests index
	
	for (( ti=0 ; ti < ${#tests[@]} ; ti++ )) ; do
		# plain test
		testscript="${tests[ti]}"
		testname="${0}/${ti}/${testscript}/plain"
		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"

		# same tests but running as function-local variables
		testscript="${ printf 'function wrapme {\n%s\n} ; wrapme\n' "${tests[ti]}" ; }"
		testname="${0}/${ti}/${testscript}/function_local"
		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"

		# same tests but running as local variables of a type discipline
		testscript="${ printf 'typeset -T wrap_t=( integer dummy ; function w {\n%s\n} ) ; wrap_t x ; x.w\n' "${tests[ti]}" ; }"
		testname="${0}/${ti}/${testscript}/type_local"
		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"

		# same tests but running as local variables of a type discipline in a namespace (this is a bit experimental)
		testscript="${ printf 'namespace ns1 { typeset -T wrap_t=( integer dummy ; function w {\n%s\n} ) ; } ; .ns1.wrap_t x ; x.w\n' "${tests[ti]}" ; }"
		testname="${0}/${ti}/${testscript}/type_local"
		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"
		[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}

# run tests
test1
test2
test3a
test3b
test4a
test4b
test_mutation1

# tests done
exit $((Errors<125?Errors:125))
