#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This test module verifies whether the AST and ksh93 builtin "getconf"
# supports the NPROCESSORS_MAX and NPROCESSORS_ONLN properties.
#
# This was reported as CR #6869888 ("AST/ksh93's "getconf" does not
# recognize "NPROCESSORS_MAX""):
#
# ------------ snip ------------
# AST/ksh93's "getconf" does not recognize "NPROCESSORS_MAX".
# The builtin "getconf" command (which is a superset of
# { /usr/bin/getconf, /usr/bin/xpg4/getconf, /usr/bin/xpg6/getconf }
# and exposes the |sysconf()| interface for scripts) already
# supports "NPROCESSORS_ONLN" and "NPROCESSORS_CONF" but
# "NPROCESSORS_MAX" is missing... ;-(
# Steps to Reproduce
#    $ getconf NPROCESSORS_MAX                                      
# Expected Result
#    <maximum number of CPUs supported>
# Actual Result
#    getconf: Invalid argument (NPROCESSORS_MAX)
# Error Message(s)
#    getconf: Invalid argument (NPROCESSORS_MAX)
# Test Case
#    $ /usr/bin/ksh93 -c 'getconf NPROCESSORS_MAX'
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="$(mktemp -t -d 'test_sun_solaris_cr_6869888_ast_getconf_does_not_recognize_NPROCESSORS_MAX.XXXXXXXX')" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function test1
{
	compound out=( typeset stdout stderr ; integer res )
	typeset key
	typeset testname
	
	for key in \
		'NPROCESSORS_CONF' \
		'NPROCESSORS_MAX' \
		'NPROCESSORS_ONLN' ; do
		testname="test1/${key}"

		out.stderr="${ { out.stdout="${ ${SHELL} -c "getconf ${key}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

		(( out.res == 0 )) || err_exit "${testname}: Test shell returned with exit code ${out.res}"
		[[ "${out.stdout}" == ~(Elr)[0-9]+ ]] || err_exit "${testname}: Expected stdout to match ~(Elr)[0-9]+, got ${ printf '%q\n' "${out.stdout}" ; }"
		[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
	done
	
	return 0
}


# run tests
test1


# cleanup
cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."


# tests done
exit $((Errors<125?Errors:125))
