#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This test module verifies whether large data passed through a subshell
# into a command substitution may trigger a hang
#
# This was reported as CR #6946957 ("Nested sub shell in command substitution with lots of data hangs"):
# ------------ snip ------------
# A command substitution will hang if the amount of output of a command
# in a sub shell exceeds a specific size.
# The important factors are:
# 1. Command substitution
# 2. sub shell
# 3. Command produces larger quantities of data
# Steps to Reproduce
#     Run ksh93 -c 'print $( ( seq 20000 ) )'
# Expected Result
#     Print numbers 1 - 20000.
# Actual Result
#     Shell hangs.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="$(mktemp -t -d 'test_sun_solaris_cr_6946957_nested_subshell_command_substitution_large_data_hangs.XXXXXXXX')" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function isvalidpid
{
        kill -0 ${1} 2>/dev/null && return 0
        return 1
}


function pretests
{
	cat <<EOF >'myseq'
#!${SHELL}

# script version of "seq"
function main
{
	integer i
	integer arg1=\$1
	integer arg2=\$2
	integer arg3=\$3

	case \$# in
		1)
			for (( i=1. ; i <= arg1 ; i=i+1. )) ; do
				printf '%d\n' i
			done
			;;
		2)
			for (( i=arg1 ; i <= arg2 ; i=i+1. )) ; do
				printf '%d\n' i
			done
			;;
		3)
			for (( i=arg1 ; i <= arg3 ; i+=arg2 )) ; do
				printf '%d\n' i
			done
			;;
		*)
			print -u2 -f '%s: Illegal number of arguments %d\n' "\$0" \$#
			return 1
			;;
	esac
	
	return 0
}

set -o noglob
IFS=''

main "\$@"
exit \$?
EOF
	chmod a+rx 'myseq'
	return 0
}

function posttests
{
	rm 'myseq'
}

# test1: enumerate some known failures with fixed tests
# (a more programmatic approach to test all iterations has been added to "sun_solaris_command_substitution.sh")
function test1
{
	# DO NOT EDIT THESE TESTS ANYMORE!!
	compound -r -a testcases=(
		# test 1a: Run test child for $(...)
		# "ulimit -c 0" is used to force the a subshell to call |fork()|
		( typeset name='test1a' cmd='print -- "$( (               myseq 60000        ) )"   ; true' )
		( typeset name='test1b' cmd='print -- "$( (               myseq 60000 ; true ) )"   ; true' )
		( typeset name='test1c' cmd='print -- "$( ( ulimit -c 0 ; myseq 60000        ) )"   ; true' )
		( typeset name='test1d' cmd='print -- "$( ( ulimit -c 0 ; myseq 60000 ; true ) )"   ; true' )

		( typeset name='test1a_nested2' cmd='print -- "$( print -- "$( (               myseq 60000        ) )" )"   ; true' )
		( typeset name='test1b_nested2' cmd='print -- "$( print -- "$( (               myseq 60000 ; true ) )" )"   ; true' )
		( typeset name='test1c_nested2' cmd='print -- "$( print -- "$( ( ulimit -c 0 ; myseq 60000        ) )" )"   ; true' )
		( typeset name='test1d_nested2' cmd='print -- "$( print -- "$( ( ulimit -c 0 ; myseq 60000 ; true ) )" )"   ; true' )

		# test 2a: Run test child for ${ ... ; }
		( typeset name='test1e' cmd='print -- "${ (               myseq 60000        ) ; }" ; true' )
		( typeset name='test1f' cmd='print -- "${ (               myseq 60000 ; true ) ; }" ; true' )
		( typeset name='test1g' cmd='print -- "${ ( ulimit -c 0 ; myseq 60000        ) ; }" ; true' )
		( typeset name='test1h' cmd='print -- "${ ( ulimit -c 0 ; myseq 60000 ; true ) ; }" ; true' )

		( typeset name='test1e_nested2' cmd='print -- "${ print -- "${ (               myseq 60000        ) ; }" ; }"   ; true' )
		( typeset name='test1f_nested2' cmd='print -- "${ print -- "${ (               myseq 60000 ; true ) ; }" ; }"   ; true' )
		( typeset name='test1g_nested2' cmd='print -- "${ print -- "${ ( ulimit -c 0 ; myseq 60000        ) ; }" ; }"   ; true' )
		( typeset name='test1h_nested2' cmd='print -- "${ print -- "${ ( ulimit -c 0 ; myseq 60000 ; true ) ; }" ; }"   ; true' )
	)
	typeset testid
	integer childpid
	integer i

	myseq 60000 >'tmpfile.expected'

	for testid in "${!testcases[@]}" ; do
		nameref currtst=testcases[testid]
				
		${SHELL} -o errexit -c "${currtst.cmd}" >'tmpfile.out' 2>'tmpfile.err' &
		(( childpid=$! ))
	
		for (( i=0 ; i < 80 ; i++ )) ; do
			isvalidpid ${childpid} || break
			sleep 0.25
		done
	
		if isvalidpid ${childpid} ; then
			err_exit "${currtst.name}: child (pid=${childpid}) still busy"
			kill -KILL ${childpid} 2>/dev/null
		fi
		wait || err_exit "${currtst.name}: Child returned non-zero exit code $?." # wait for child (and/or avoid zombies/slime)
	
		[[ "$( < 'tmpfile.err' )" == '' ]] || err_exit "${currtst.name}: Expected empty stderr, got ${ printf '%q\n' "$( < 'tmpfile.err' )" ; }"

		# compare input/output
		cmp -s 'tmpfile.expected' 'tmpfile.out' || err_exit "${currtst.name}: tmpfile.expected and tmpfile.out differ"

		rm 'tmpfile.out' 'tmpfile.err'
	done
	
	rm 'tmpfile.expected' 

	return 0
}


# prepare for tests
export PATH="${PWD}:${PATH}"
builtin cmp
builtin rm
unset VMALLOC_OPTIONS VMDEBUG # no debug for helper scripts

pretests

# run tests
test1

# cleanup
posttests

cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
