#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2010, 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This test module verifies that "tail -0f" prints no lines before
# following a file.
#
# This was reported as CR #6949852 ("tail -0f outputs the entire file"):
# ------------ snip ------------
# With a recent ON (04/29/2010) ksh93 will use the builtin for "tail".
# It appears to be printing the entire file when given the options "-0f".
# i.e.
# 
# # which tail
# /usr/bin/tail
# # tail -0f /tmp/test                                                        
# line 1
# line 2
# line 3
# line 4
# line 5
# line 6
# line 7
# line 8
# line 9
# line 10
# line 11
# ^C
# # /usr/bin/tail -0f /tmp/test
# ^C
# #
# 
# tail -1f works fine:
# 
# # tail -1f /tmp/test
# line 11
# ^C                                               
# 
# /usr/bin/tail, /usr/xpg4/bin/tail both do the right thing.
# The built-in gets it wrong.
# 
# This may be a problem for shell scripts as /bin/sh is ksh93 now.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_cr_6949852_tail_0f_outputs_entire_file.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function test_tail_0f
{
	typeset -r cmd="${1}"
	typeset -r testfile='testfile1.txt'
	typeset -r child_stdout='out_stdout.txt'
	typeset -r child_stderr='out_stderr.txt'
	integer childpid
	typeset testname
	typeset out
	integer vi

	compound -r -a variants=(
		( typeset -r -a args=( '-0f' ) )
		( typeset -r -a args=( '-0' '-f' ) )
		( typeset -r -a args=( '-n' '0' '-f' ) )
	)

	for (( vi=0 ; vi < ${#variants[@]} ; vi++ )) ; do
		testname="${0}/${cmd},${vi}"
		
		print $'one\ntwo\nthree' >"${testfile}"
	
		# fixme: it should be [vi], not [${vi}]
		${cmd} "${variants[${vi}].args[@]}" "${testfile}" 2>"${child_stderr}" >"${child_stdout}" &
		(( childpid=$! ))
		
		sleep 2
	
		out="$( <"${child_stdout}" )"
		[[ "${out}" == '' ]] || err_exit "${testname}: Expected empty stdout, got ${ printf '%q\n' "${out}" ; }"
		
		print 'four' >>"${testfile}"
		sleep 4
		
		out="$( <"${child_stdout}" )"; [[ "${out}" == 'four' ]] || err_exit "${testname}: Expected stdout=='four', got ${ printf '%q\n' "${out}" ; }"
		
		kill -KILL ${childpid}
		wait
	
		out="$( <"${child_stderr}" )" ; [[ "${out}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out}" ; }"
		
		rm "${testfile}" "${child_stdout}" "${child_stderr}"
	done

	return 0
}

# prepare
builtin tail

# run tests
# fixme: Test /usr/bin/tail once we use AST "tail" for it.
for cmd in \
	'tail' \
	'/usr/xpg4/bin/tail' ; do
	test_tail_0f "${cmd}"
done

cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."


# tests done
exit $((Errors<125?Errors:125))
