#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This module tests whether the libcmd tee(1) builtin utility properly
# handles the case when a child process terminates while the tee
# builtin is in an |open()| syscall.
#
# This was reported as CR #6997060 ("tee(1) fails to open file with
# EINTR when used in a pipeline"):
# ------------ snip ------------
# nightly using NFS often fails with the following error message when
# lint'ing under cmd/sgs:
# 
# tee: lint.out: cannot create [Interrupted system call]
# 
# while doing "cat lint.out.64 | tee -a lint.out". Same failure can be
# seen with the following script:
# 
# $ cat A
# while :
# do
#         echo "a" | tee $1 > /dev/null
#         if [ $? -ne 0 ]; then
#                 exit 1
#         fi
# done
# $ ksh ./A ~/out
# tee: /home/nakanon/out: cannot create [Interrupted system call]
# 
# ~/out is on the NFS mounted directory. If this doesn't reproduce the
# failure, use this:
# 
# mknod pipe p
# (sleep 1; cat pipe) &
# (sleep 0.5; echo aa) | tee -a pipe
# rm pipe
# 
# This should fail with the same error message.
# ksh88, bash, xpg4sh doesn't fail.
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_cr_6997060_tee_fails_open_with_EINTR.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function test1
{
	typeset -r testscript=$'
		builtin tee
		/usr/sbin/mknod "pipe" p
		(sleep 4 ; cat "pipe") &
		(sleep 2 ; print "aa") | tee -a "pipe"
		rm "pipe"'
	typeset -r expected_output=$'aa\naa'
	compound out=( typeset stdout stderr ; integer res )
	
	out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -c "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

	[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${0}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
	[[ "${out.stderr}" == '' ]] || err_exit "${0}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
	(( out.res == 0 )) || err_exit "${0}: Unexpected exit code ${out.res}"

	return 0
}


# run tests
test1


cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
